# Flag: Always write data
NORMAL = 0

# Flag: Only write if the verbosity is "verbose" or greater.
VERBOSE = 1

# Flag: Only write if the verbosity is "very verbose" or greater.
VERY_VERBOSE = 2

# Flag: Only write if the verbosity is "debug".
DEBUG = 4
